@@ -1,5 +1,5 @@
 /*
- * $Id: filedir.c,v 1.32.2.3 2003/06/09 14:53:15 srittau Exp $
+ * $Id: filedir.c,v 1.45.2.2.2.14 2004/10/06 20:05:14 bfernhomberg Exp $
  *
  * Copyright (c) 1990,1993 Regents of The University of Michigan.
  * All Rights Reserved.  See COPYRIGHT.
@@ -9,25 +9,8 @@
 #include "config.h"
 #endif /* HAVE_CONFIG_H */
 
-#include <errno.h>
-#include <atalk/logger.h>
-#include <sys/types.h>
-#include <sys/stat.h>
-#include <sys/param.h>
-#include <netatalk/endian.h>
-#include <atalk/adouble.h>
-#include <atalk/afp.h>
-#include <atalk/util.h>
-#ifdef CNID_DB
-#include <atalk/cnid.h>
-#endif /* CNID_DB */
 #include <stdio.h>
 #include <stdlib.h>
-#ifdef HAVE_FCNTL_H
-#include <fcntl.h>
-#endif /* HAVE_FCNTL_H */
-#include <dirent.h>
-
 /* STDC check */
 #if STDC_HEADERS
 #include <string.h>
@@ -43,9 +26,18 @@ char *strchr (), *strrchr ();
 #endif /* ! HAVE_MEMCPY */
 #endif /* STDC_HEADERS */
 
-#ifdef HAVE_UNISTD_H
-#include <unistd.h>
-#endif /* HAVE_UNISTD_H */
+#ifdef HAVE_STRINGS_H
+#include <strings.h>
+#endif
+#include <dirent.h>
+#include <errno.h>
+#include <sys/param.h>
+#include <atalk/adouble.h>
+
+#include <atalk/afp.h>
+#include <atalk/util.h>
+#include <atalk/cnid.h>
+#include <atalk/logger.h>
 
 #include "directory.h"
 #include "desktop.h"
@@ -56,6 +48,7 @@ char *strchr (), *strrchr ();
 #include "filedir.h"
 #include "unix.h"
 
+#ifdef DROPKLUDGE
 int matchfile2dirperms(upath, vol, did)
 /* Since it's kinda' big; I decided against an
 inline function */
@@ -69,7 +62,7 @@ more information */
     struct stat	st, sb;
     struct dir	*dir;
     char	*adpath;
-    int		uid;
+    uid_t       uid;
     int         ret = AFP_OK;
 #ifdef DEBUG
     LOG(log_info, logtype_afpd, "begin matchfile2dirperms:");
@@ -80,7 +73,7 @@ more information */
         return AFPERR_NOOBJ ;
     }
 
-    adpath = ad_path( upath, ADFLAGS_HF );
+    adpath = vol->ad_path( upath, ADFLAGS_HF );
     /* FIXME dirsearch doesn't move cwd to did ! */
     if (( dir = dirlookup( vol, did )) == NULL ) {
         LOG(log_error, logtype_afpd, "matchfile2dirperms: Unable to get directory info.");
@@ -89,7 +82,7 @@ more information */
     else if (stat(".", &sb) < 0) {
         LOG(log_error, logtype_afpd,
             "matchfile2dirperms: Error checking directory \"%s\": %s",
-            dir->d_name, strerror(errno));
+            dir->d_m_name, strerror(errno));
         ret = AFPERR_NOOBJ;
     }
     else {
@@ -100,29 +93,29 @@ more information */
             if (lchown(upath, sb.st_uid, sb.st_gid) < 0)
             {
                 LOG(log_error, logtype_afpd,
-                    "matchfile2dirperms: Error changing owner/gid of %s: %s",
+                    "matchfile2dirperms(%s): Error changing owner/gid: %s",
                     upath, strerror(errno));
                 ret = AFPERR_ACCESS;
             }
             else if (chmod(upath,(st.st_mode&~default_options.umask)| S_IRGRP| S_IROTH) < 0)
             {
                 LOG(log_error, logtype_afpd,
-                    "matchfile2dirperms:  Error adding file read permissions: %s",
-                    strerror(errno));
+                    "matchfile2dirperms(%s): Error adding file read permissions: %s",
+                    upath, strerror(errno));
                 ret = AFPERR_ACCESS;
             }
             else if (lchown(adpath, sb.st_uid, sb.st_gid) < 0)
             {
                 LOG(log_error, logtype_afpd,
-                    "matchfile2dirperms: Error changing AppleDouble owner/gid %s: %s",
+                    "matchfile2dirperms(%s): Error changing AppleDouble owner/gid: %s",
                     adpath, strerror(errno));
                 ret = AFPERR_ACCESS;
             }
             else if (chmod(adpath, (st.st_mode&~default_options.umask)| S_IRGRP| S_IROTH) < 0)
             {
                 LOG(log_error, logtype_afpd,
-                    "matchfile2dirperms:  Error adding AD file read permissions: %s",
-                    strerror(errno));
+                    "matchfile2dirperms(%s):  Error adding AD file read permissions: %s",
+                    adpath, strerror(errno));
                 ret = AFPERR_ACCESS;
             }
             seteuid(uid); 
@@ -134,20 +127,20 @@ more information */
 #endif /* DEBUG */
     return ret;
 }
-
+#endif
 
 int afp_getfildirparams(obj, ibuf, ibuflen, rbuf, rbuflen )
 AFPObj      *obj;
 char	*ibuf, *rbuf;
 int		ibuflen, *rbuflen;
 {
-    struct stat		st;
+    struct stat		*st;
     struct vol		*vol;
     struct dir		*dir;
     u_int32_t           did;
     int			buflen, ret;
-    char		*path;
     u_int16_t		fbitmap, dbitmap, vid;
+    struct path         *s_path;
 
 #ifdef DEBUG
     LOG(log_info, logtype_afpd, "begin afp_getfildirparams:");
@@ -158,16 +151,15 @@ int		ibuflen, *rbuflen;
 
     memcpy( &vid, ibuf, sizeof( vid ));
     ibuf += sizeof( vid );
-    if (( vol = getvolbyvid( vid )) == NULL ) {
+    if (NULL == ( vol = getvolbyvid( vid )) ) {
         return( AFPERR_PARAM );
     }
 
     memcpy( &did, ibuf, sizeof( did ));
     ibuf += sizeof( did );
 
-    if (( dir = dirlookup( vol, did )) == NULL ) {
-        /* FIXME: EACCES */
-        return( AFPERR_NOOBJ );
+    if (NULL == ( dir = dirlookup( vol, did )) ) {
+        return afp_errno;
     }
 
     memcpy( &fbitmap, ibuf, sizeof( fbitmap ));
@@ -177,35 +169,40 @@ int		ibuflen, *rbuflen;
     dbitmap = ntohs( dbitmap );
     ibuf += sizeof( dbitmap );
 
-    if (( path = cname( vol, dir, &ibuf )) == NULL || 
-          stat( mtoupath(vol, path ), &st ) < 0 ) {
-        switch (errno) {
-        case EACCES:
-        case EPERM:
-            return AFPERR_ACCESS;
-        default:
-            return AFPERR_NOOBJ;
-        }
+    if (NULL == ( s_path = cname( vol, dir, &ibuf )) ) {
+        return get_afp_errno(AFPERR_NOOBJ); 
+    }
+
+    st   = &s_path->st;
+    if (!s_path->st_valid) {
+        /* it's a dir and it should be there
+         * because we chdir in it in cname or
+         * it's curdir (maybe deleted, but then we can't know).
+         * So we need to try harder.
+         */
+        of_statdir(vol, s_path);
+    }
+    if ( s_path->st_errno != 0 ) {
+        return( AFPERR_NOOBJ );
     }
 
     buflen = 0;
-    if (S_ISDIR(st.st_mode)) {
+    if (S_ISDIR(st->st_mode)) {
         if (dbitmap) {
-            if (*path != '\0') {
-            	/* the dir wasn't in the cache and we weren't able to chdir in it.
-            	*/
-            	return AFPERR_ACCESS;
-            }
-            ret = getdirparams(vol, dbitmap, ".", curdir,
-                               &st, rbuf + 3 * sizeof( u_int16_t ), &buflen );
+            dir = s_path->d_dir;
+            if (!dir) 
+                return AFPERR_NOOBJ;
+
+            ret = getdirparams(vol, dbitmap, s_path, dir,
+                                 rbuf + 3 * sizeof( u_int16_t ), &buflen );
             if (ret != AFP_OK )
                 return( ret );
         }
         /* this is a directory */
         *(rbuf + 2 * sizeof( u_int16_t )) = (char) FILDIRBIT_ISDIR;
     } else {
-        if (fbitmap && ( ret = getfilparams(vol, fbitmap, path, curdir, &st,
-                                            rbuf + 3 * sizeof( u_int16_t ), &buflen )) != AFP_OK ) {
+        if (fbitmap && AFP_OK != (ret = getfilparams(vol, fbitmap, s_path, curdir, 
+                                            rbuf + 3 * sizeof( u_int16_t ), &buflen )) ) {
             return( ret );
         }
         /* this is a file */
@@ -227,41 +224,15 @@ int		ibuflen, *rbuflen;
     return( AFP_OK );
 }
 
-/*
- * We can't use unix file's perm to support Apple's inherited protection modes.
- * If we aren't the file's owner we can't change its perms when moving it and smb
- * nfs,... don't even try.
-*/
-#define AFP_CHECK_ACCESS 
-
-int check_access(char *path, int mode)
-{
-#ifdef AFP_CHECK_ACCESS
-struct maccess ma;
-char *p;
-
-    p = ad_dir(path);
-    if (!p)
-       return -1;
-
-    accessmode(p, &ma, curdir, NULL);
-    if ((mode & OPENACC_WR) && !(ma.ma_user & AR_UWRITE))
-        return -1;
-    if ((mode & OPENACC_RD) && !(ma.ma_user & AR_UREAD))
-        return -1;
-#endif
-    return 0;
-}
-
 int afp_setfildirparams(obj, ibuf, ibuflen, rbuf, rbuflen )
 AFPObj      *obj;
 char	*ibuf, *rbuf;
 int		ibuflen, *rbuflen;
 {
-    struct stat	st;
+    struct stat	*st;
     struct vol	*vol;
     struct dir	*dir;
-    char	*path;
+    struct path *path;
     u_int16_t	vid, bitmap;
     int		did, rc;
 
@@ -274,7 +245,7 @@ int		ibuflen, *rbuflen;
     memcpy( &vid, ibuf, sizeof(vid));
     ibuf += sizeof( vid );
 
-    if (( vol = getvolbyvid( vid )) == NULL ) {
+    if (NULL == ( vol = getvolbyvid( vid )) ) {
         return( AFPERR_PARAM );
     }
 
@@ -284,22 +255,29 @@ int		ibuflen, *rbuflen;
     memcpy( &did, ibuf, sizeof( did));
     ibuf += sizeof( did);
 
-    if (( dir = dirlookup( vol, did )) == NULL ) {
-        return( AFPERR_NOOBJ );
+    if (NULL == ( dir = dirlookup( vol, did )) ) {
+	return afp_errno;    
     }
 
     memcpy( &bitmap, ibuf, sizeof( bitmap ));
     bitmap = ntohs( bitmap );
     ibuf += sizeof( bitmap );
 
-    if (( path = cname( vol, dir, &ibuf )) == NULL ) {
-        return( AFPERR_NOOBJ );
+    if (NULL == ( path = cname( vol, dir, &ibuf ))) {
+        return get_afp_errno(AFPERR_NOOBJ); 
     }
 
-    if ( stat( mtoupath(vol, path ), &st ) < 0 ) {
-        return( AFPERR_NOOBJ );
+    st   = &path->st;
+    if (!path->st_valid) {
+        /* it's a dir and it should be there
+         * because we chdir in it in cname
+         */
+        of_statdir(vol, path);
     }
 
+    if ( path->st_errno != 0 ) {
+        return( AFPERR_NOOBJ );
+    }
     /*
      * If ibuf is odd, make it even.
      */
@@ -307,7 +285,7 @@ int		ibuflen, *rbuflen;
         ibuf++;
     }
 
-    if (S_ISDIR(st.st_mode)) {
+    if (S_ISDIR(st->st_mode)) {
         rc = setdirparams(vol, path, bitmap, ibuf );
     } else {
         rc = setfilparams(vol, path, bitmap, ibuf );
@@ -324,7 +302,7 @@ int		ibuflen, *rbuflen;
 }
 
 /* -------------------------------------------- 
-   Factorise some check on a pathname
+   Factorise some checks on a pathname
 */
 int check_name(const struct vol *vol, char *name)
 {
@@ -335,7 +313,7 @@ int check_name(const struct vol *vol, char *name)
     if ((vol->v_flags & AFPVOL_NOHEX) && strchr(name, '/'))
         return AFPERR_PARAM;
 
-    if (!validupath(vol, name))
+    if (!vol->validupath(vol, name))
         return AFPERR_EXIST;
 
     /* check for vetoed filenames */
@@ -359,36 +337,41 @@ int         isdir;
     char            *p;
     char            *upath;
     int             rc;
-    struct stat     st;
+    struct stat     *st, nst;
     int             adflags;
     struct adouble	ad;
     struct adouble	*adp;
-    struct ofork	*opened;
-
-#ifdef CNID_DB
+    struct ofork	*opened = NULL;
+    struct path         path;
     cnid_t      id;
-#endif /* CNID_DB */
 
-    memset(&ad, 0, sizeof(ad));
+    ad_init(&ad, vol->v_adouble);
     adp = &ad;
     adflags = 0;
     
     if (!isdir) {
-#ifdef CNID_DB
-        p = mtoupath(vol, oldname);
-        id = cnid_get(vol->v_db, sdir->d_did, p, strlen(p));
-#endif /* CNID_DB */
+        p = mtoupath(vol, oldname, sdir->d_did, utf8_encoding());
+        if (!p) { 
+            return AFPERR_PARAM; /* can't convert */
+        }
+        id = cnid_get(vol->v_cdb, sdir->d_did, p, strlen(p));
         p = ctoupath( vol, sdir, oldname );
-        if ((opened = of_findname(p, NULL))) {
+        if (!p) { 
+            return AFPERR_PARAM; /* pathname too long */
+        }
+        path.st_valid = 0;
+        path.u_name = p;
+        if ((opened = of_findname(&path))) {
             /* reuse struct adouble so it won't break locks */
             adp = opened->of_ad;
         }
     }
     else {
-#ifdef CNID_DB
         id = sdir->d_did; /* we already have the CNID */
-#endif /* CNID_DB */
         p = ctoupath( vol, sdir->d_parent, oldname );
+        if (!p) {
+            return AFPERR_PARAM;
+        }
         adflags = ADFLAGS_DIR;
     }
     /*
@@ -397,7 +380,7 @@ int         isdir;
      * we are in the dest folder so we need to use p for ad_open
     */
     
-    if (!ad_open(p, ADFLAGS_HF |adflags, O_RDONLY, 0666, adp)) {
+    if (!ad_metadata(p, adflags, adp)) {
     u_int16_t bshort;
 
         ad_getattr(adp, &bshort);
@@ -406,43 +389,50 @@ int         isdir;
             return(AFPERR_OLOCK);
     }
 
-    upath = mtoupath(vol, newname);
+    if (NULL == (upath = mtoupath(vol, newname, curdir->d_did, utf8_encoding()))){ 
+        return AFPERR_PARAM;
+    }
+    path.u_name = upath;
+    st = &path.st;    
     if (0 != (rc = check_name(vol, upath))) {
             return  rc;
     }
 
     /* source == destination. we just silently accept this. */
-    if (curdir == sdir) {
+    if ((!isdir && curdir == sdir) || (isdir && curdir == sdir->d_parent)) {
         if (strcmp(oldname, newname) == 0)
             return AFP_OK;
 
-        /* deal with case insensitive, case-preserving filesystems. */
-        if ((stat(upath, &st) == 0) && strdiacasecmp(oldname, newname))
-            return AFPERR_EXIST;
-
-    } else if (stat(upath, &st ) == 0)
+        if (stat(upath, st) == 0) {
+            if (!stat(p, &nst) && !(nst.st_dev == st->st_dev && nst.st_ino == st->st_ino) ) {
+                /* not the same file */
+                return AFPERR_EXIST;
+            }
+            errno = 0;
+        }
+    } else if (stat(upath, st ) == 0)
         return AFPERR_EXIST;
 
     if ( !isdir ) {
-        if (of_findname(upath, &st)) {
+        path.st_valid = 1;
+        path.st_errno = errno;
+        if (of_findname(&path)) {
             rc = AFPERR_EXIST; /* was AFPERR_BUSY; */
         } else {
-            rc = renamefile( p, upath, newname,vol_noadouble(vol), adp );
+            rc = renamefile(vol, p, upath, newname, adp );
             if (rc == AFP_OK)
                 of_rename(vol, opened, sdir, oldname, curdir, newname);
         }
     } else {
-        rc = renamedir(p, upath, sdir, curdir, newname, vol_noadouble(vol));
+        rc = renamedir(vol, p, upath, sdir, curdir, newname);
     }
-    if ( rc == AFP_OK ) {
-#ifdef CNID_DB
+    if ( rc == AFP_OK && id ) {
         /* renaming may have moved the file/dir across a filesystem */
-        if (stat(upath, &st) < 0)
+        if (stat(upath, st) < 0)
             return AFPERR_MISC;
 
         /* fix up the catalog entry */
-        cnid_update(vol->v_db, id, &st, curdir->d_did, upath, strlen(upath));
-#endif /* CNID_DB */
+        cnid_update(vol->v_cdb, id, st, curdir->d_did, upath, strlen(upath));
     }
 
     return rc;
@@ -456,9 +446,10 @@ int		ibuflen, *rbuflen;
 {
     struct vol	*vol;
     struct dir	*sdir;
-    char		*path, *oldname, *newname;
+    char        *oldname, *newname;
+    struct path *path;
     u_int32_t	did;
-    int			plen;
+    int         plen;
     u_int16_t	vid;
     int         isdir = 0;
     int         rc;
@@ -471,7 +462,7 @@ int		ibuflen, *rbuflen;
 
     memcpy( &vid, ibuf, sizeof( vid ));
     ibuf += sizeof( vid );
-    if (( vol = getvolbyvid( vid )) == NULL ) {
+    if (NULL == ( vol = getvolbyvid( vid )) ) {
         return( AFPERR_PARAM );
     }
 
@@ -480,20 +471,25 @@ int		ibuflen, *rbuflen;
 
     memcpy( &did, ibuf, sizeof( did ));
     ibuf += sizeof( did );
-    if (( sdir = dirlookup( vol, did )) == NULL ) {
-        return( AFPERR_NOOBJ );
+    if (NULL == ( sdir = dirlookup( vol, did )) ) {
+	return afp_errno;    
     }
 
     /* source pathname */
-    if (( path = cname( vol, sdir, &ibuf )) == NULL ) {
-        return( AFPERR_NOOBJ );
+    if (NULL == ( path = cname( vol, sdir, &ibuf )) ) {
+        return get_afp_errno(AFPERR_NOOBJ); 
     }
 
     sdir = curdir;
     newname = obj->newtmp;
     oldname = obj->oldtmp;
-    if ( *path != '\0' ) {
-        strcpy(oldname, path); /* an extra copy for of_rename */
+    isdir = path_isadir(path);
+    if ( *path->m_name != '\0' ) {
+        strcpy(oldname, path->m_name); /* an extra copy for of_rename */
+        if (isdir) {
+            /* curdir parent dir, need to move sdir back */
+            sdir = path->d_dir;
+        }
     }
     else {
         if ( sdir->d_parent == NULL ) { /* root directory */
@@ -501,25 +497,17 @@ int		ibuflen, *rbuflen;
         }
         /* move to destination dir */
         if ( movecwd( vol, sdir->d_parent ) < 0 ) {
-            return( AFPERR_NOOBJ );
+            return afp_errno;
         }
-        isdir = 1;
-        strcpy(oldname, sdir->d_name);
+        strcpy(oldname, sdir->d_m_name);
     }
 
     /* another place where we know about the path type */
-    if ( *ibuf++ != 2 ) {
+    if ((plen = copy_path_name(newname, ibuf)) < 0) {
         return( AFPERR_PARAM );
     }
 
-    if (( plen = (unsigned char)*ibuf++ ) != 0 ) {
-        strncpy( newname, ibuf, plen );
-        newname[ plen ] = '\0';
-        if (strlen(newname) != plen) {
-            return( AFPERR_PARAM );
-        }
-    }
-    else {
+    if (!plen) {
         return AFP_OK; /* newname == oldname same dir */
     }
     
@@ -544,7 +532,8 @@ int		ibuflen, *rbuflen;
 {
     struct vol		*vol;
     struct dir		*dir;
-    char		*path, *upath;
+    struct path         *s_path;
+    char		*upath;
     int			did, rc;
     u_int16_t		vid;
 
@@ -557,7 +546,7 @@ int		ibuflen, *rbuflen;
 
     memcpy( &vid, ibuf, sizeof( vid ));
     ibuf += sizeof( vid );
-    if (( vol = getvolbyvid( vid )) == NULL ) {
+    if (NULL == ( vol = getvolbyvid( vid )) ) {
         return( AFPERR_PARAM );
     }
 
@@ -566,26 +555,29 @@ int		ibuflen, *rbuflen;
 
     memcpy( &did, ibuf, sizeof( did ));
     ibuf += sizeof( int );
-    if (( dir = dirlookup( vol, did )) == NULL ) {
-        return( AFPERR_NOOBJ );
+    if (NULL == ( dir = dirlookup( vol, did )) ) {
+	return afp_errno;    
     }
 
-    if (( path = cname( vol, dir, &ibuf )) == NULL ) {
-        return( AFPERR_NOOBJ );
+    if (NULL == ( s_path = cname( vol, dir, &ibuf )) ) {
+        return get_afp_errno(AFPERR_NOOBJ); 
     }
 
-    upath = mtoupath(vol, path );
-    if ( *path == '\0' ) {
-        rc = deletecurdir( vol, obj->oldtmp, AFPOBJ_TMPSIZ);
-    } else if (of_findname(upath, NULL)) {
+    upath = s_path->u_name;
+    if ( path_isadir( s_path) ) {
+    	if (*s_path->m_name != '\0') {
+    	    rc = AFPERR_ACCESS;
+    	}
+    	else {
+            rc = deletecurdir( vol, obj->oldtmp, AFPOBJ_TMPSIZ);
+        }
+    } else if (of_findname(s_path)) {
         rc = AFPERR_BUSY;
-    } else if ((rc = deletefile( upath, 1)) == AFP_OK) {
-#ifdef CNID_DB /* get rid of entry */
-        cnid_t id = cnid_get(vol->v_db, curdir->d_did, upath, strlen(upath));
-        cnid_delete(vol->v_db, id);
-#endif /* CNID_DB */
+    } else {
+        rc = deletefile(vol, upath, 1);
     }
     if ( rc == AFP_OK ) {
+	curdir->offcnt--;
         setvoltime(obj, vol );
     }
 
@@ -595,36 +587,58 @@ int		ibuflen, *rbuflen;
 
     return( rc );
 }
+/* ------------------------ */
+char *absupath( vol, dir, u )
+const struct vol	*vol;
+struct dir	*dir;
+char	*u;
+{
+    struct dir	*d;
+    static char	path[ MAXPATHLEN + 1];
+    char	*p;
+    int		len;
 
+    if (u == NULL)
+        return NULL;
+        
+    p = path + sizeof( path ) - 1;
+    *p = '\0';
+    len = strlen( u );
+    p -= len;
+    memcpy( p, u, len );
+    if (dir) for ( d = dir; d->d_parent; d = d->d_parent ) {
+        u = d->d_u_name;
+        len = strlen( u );
+        if (p -len -1 < path) {
+            /* FIXME 
+               rather rare so LOG error and/or client message ?
+            */
+            return NULL;
+        }
+        *--p = '/';
+        p -= len;
+        memcpy( p, u, len );
+    }
+    len = strlen( vol->v_path );
+    if (p -len -1 < path) {
+        return NULL;
+    }
+    *--p = '/';
+    p -= len;
+    memcpy( p, vol->v_path, len );
+
+    return( p );
+}
+
+/* ------------------------
+ * FIXME dir could be NULL
+*/
 char *ctoupath( vol, dir, name )
 const struct vol	*vol;
 struct dir	*dir;
 char	*name;
 {
-    struct dir	*d;
-    static char	path[ MAXPATHLEN + 1];
-    char	*p, *u;
-    int		len;
-
-    p = path + sizeof( path ) - 1;
-    *p = '\0';
-    u = mtoupath(vol, name );
-    len = strlen( u );
-    p -= len;
-    strncpy( p, u, len );
-    for ( d = dir; d->d_parent; d = d->d_parent ) {
-        *--p = '/';
-        u = mtoupath(vol, d->d_name );
-        len = strlen( u );
-        p -= len;
-        strncpy( p, u, len );
-    }
-    *--p = '/';
-    len = strlen( vol->v_path );
-    p -= len;
-    strncpy( p, vol->v_path, len );
-
-    return( p );
+    return absupath(vol, dir, mtoupath(vol, name, dir->d_did, utf8_encoding()));
 }
 
 /* ------------------------- */
@@ -635,11 +649,12 @@ int		ibuflen, *rbuflen;
 {
     struct vol	*vol;
     struct dir	*sdir, *ddir;
-    int         isdir = 0;
-    char	    *oldname, *newname;
-    char        *path;
+    int         isdir;
+    char	*oldname, *newname;
+    struct path *path;
     int		did;
-    int		plen;
+    int		pdid;
+    int         plen;
     u_int16_t	vid;
     int         rc;
 #ifdef DROPKLUDGE
@@ -655,7 +670,7 @@ int		ibuflen, *rbuflen;
 
     memcpy( &vid, ibuf, sizeof( vid ));
     ibuf += sizeof( vid );
-    if (( vol = getvolbyvid( vid )) == NULL ) {
+    if (NULL == ( vol = getvolbyvid( vid )) ) {
         return( AFPERR_PARAM );
     }
 
@@ -665,73 +680,78 @@ int		ibuflen, *rbuflen;
     /* source did followed by dest did */
     memcpy( &did, ibuf, sizeof( did ));
     ibuf += sizeof( int );
-    if (( sdir = dirlookup( vol, did )) == NULL ) {
-        return( AFPERR_PARAM );
+    if (NULL == ( sdir = dirlookup( vol, did )) ) {
+        return afp_errno; /* was AFPERR_PARAM */
     }
 
     memcpy( &did, ibuf, sizeof( did ));
     ibuf += sizeof( int );
 
     /* source pathname */
-    if (( path = cname( vol, sdir, &ibuf )) == NULL ) {
-        return( AFPERR_NOOBJ );
+    if (NULL == ( path = cname( vol, sdir, &ibuf )) ) {
+        return get_afp_errno(AFPERR_NOOBJ); 
     }
 
     sdir = curdir;
     newname = obj->newtmp;
     oldname = obj->oldtmp;
-    if ( *path != '\0' ) {
-        /* not a directory */
-        strcpy(oldname, path); /* an extra copy for of_rename */
+    
+    isdir = path_isadir(path);
+    if ( *path->m_name != '\0' ) {
+        if (isdir) {
+            sdir = path->d_dir;
+	}
+        strcpy(oldname, path->m_name); /* an extra copy for of_rename */
     } else {
-        isdir = 1;
-        strcpy(oldname, sdir->d_name);
+        strcpy(oldname, sdir->d_m_name);
     }
 
     /* get the destination directory */
-    if (( ddir = dirlookup( vol, did )) == NULL ) {
-        return( AFPERR_PARAM );
+    if (NULL == ( ddir = dirlookup( vol, did )) ) {
+        return afp_errno; /*  was AFPERR_PARAM */
     }
-    if (( path = cname( vol, ddir, &ibuf )) == NULL ) {
+    if (NULL == ( path = cname( vol, ddir, &ibuf ))) {
         return( AFPERR_NOOBJ );
     }
-    if ( *path != '\0' ) {
-        return( AFPERR_BADTYPE );
+    pdid = curdir->d_did;
+    if ( *path->m_name != '\0' ) {
+        return path_error(path, AFPERR_NOOBJ);
     }
 
     /* one more place where we know about path type */
-    if ( *ibuf++ != 2 ) {
+    if ((plen = copy_path_name(newname, ibuf)) < 0) {
         return( AFPERR_PARAM );
     }
 
-    if (( plen = (unsigned char)*ibuf++ ) != 0 ) {
-        strncpy( newname, ibuf, plen );
-        newname[ plen ] = '\0';
-        if (strlen(newname) != plen) {
-            return( AFPERR_PARAM );
-        }
-    }
-    else {
+    if (!plen) {
         strcpy(newname, oldname);
     }
-    
+
     rc = moveandrename(vol, sdir, oldname, newname, isdir);
 
     if ( rc == AFP_OK ) {
-        char *upath = mtoupath(vol, newname);
+        char *upath = mtoupath(vol, newname, pdid, utf8_encoding());
+        
+        if (NULL == upath) {
+            return AFPERR_PARAM;
+        }
+        curdir->offcnt++;
+        sdir->offcnt--;
 #ifdef DROPKLUDGE
         if (vol->v_flags & AFPVOL_DROPBOX) {
-            if (retvalue=matchfile2dirperms (upath, vol, did) != AFP_OK) {
+            /* FIXME did is not always the source id */
+            if ((retvalue=matchfile2dirperms (upath, vol, did)) != AFP_OK) {
                 return retvalue;
             }
         }
         else
 #endif /* DROPKLUDGE */
-            if (!isdir) {
+            /* if unix priv don't try to match perm with dest folder */
+            if (!isdir && !vol_unix_priv(vol)) {
                 int  admode = ad_mode("", 0777);
 
                 setfilmode(upath, admode, NULL);
-                setfilmode(ad_path( upath, ADFLAGS_HF ), ad_hf_mode(admode), NULL);
+                setfilmode(vol->ad_path( upath, ADFLAGS_HF ), ad_hf_mode(admode), NULL);
             }
         setvoltime(obj, vol );
     }
