@@ -50,13 +50,14 @@
 
 #define COUNT_ARRAY(array) (sizeof((array))/sizeof((array)[0]))
 #define NUMOF COUNT_ARRAY
-#undef  KEEP_LOGFILES_OPEN
+#define KEEP_LOGFILES_OPEN
 #define DO_SYSLOG
 #define DO_FILELOG
 
 #undef  DEBUG_OUTPUT_TO_SCREEN
 #undef  CHECK_STAT_ON_NEW_FILES 
 #undef  CHECK_ACCESS_ON_NEW_FILES 
+#define OPEN_LOGS_AS_UID 0
 
 /* =========================================================================
     External function declarations
@@ -234,16 +235,18 @@ bool log_setup(char *filename, enum loglevels loglevel, enum logtypes logtype,
 {
 #ifndef DISABLE_LOGGER
 
+#ifdef CHECK_STAT_ON_NEW_FILES
   struct stat statbuf;
   int firstattempt;
   int retval;
   gid_t gid;
   uid_t uid;
   int access;
+#endif
   char lastchar[2];
 
   log_file_data_pair *logs;
-
+  
   log_init();
 
   logs = log_file_arr[logtype];
@@ -372,6 +375,7 @@ bool log_setup(char *filename, enum loglevels loglevel, enum logtypes logtype,
   }
 #endif /* CHECK_ACCESS_ON_NEW_FILES */
 #endif /* CHECK_STAT_ON_NEW_FILES */
+/*
 #ifdef KEEP_LOGFILES_OPEN
   if ((*logs)[1].log_file!=NULL)
     fclose((*logs)[1].log_file);
@@ -385,6 +389,7 @@ bool log_setup(char *filename, enum loglevels loglevel, enum logtypes logtype,
     return false;
   }
 #endif
+*/
 
   LOG(log_debug7, logtype_logger, "log_file_arr[%d] now contains: "
 	                          "{log_filename:%s, log_file:%p, log_level: %d}", logtype,
@@ -508,6 +513,81 @@ make_log_func set_log_location(char *srcfilename, int srclinenumber)
 }
 #endif /* DISABLE_LOGGER */
 
+/* ---------------------- 
+ * chainsawed from ethereal
+*/
+#ifdef isprint
+#undef isprint
+#endif
+#define isprint(c) ((c) >= 0x20 && (c) < 0x7f)
+
+static char *format_text(char *fmtbuf, const char *string)
+{
+  int column;
+  const char *stringend = string + strlen(string);
+  char c;
+  int i;
+
+  column = 0;
+  while (string < stringend) {
+    c = *string++;
+
+    if (isprint(c)) {
+      fmtbuf[column] = c;
+      column++;
+    } else {
+      fmtbuf[column] =  '\\';
+      column++;
+      switch (c) {
+      case '\\':
+	fmtbuf[column] = '\\';
+	column++;
+	break;
+      case '\a':
+	fmtbuf[column] = 'a';
+	column++;
+	break;
+      case '\b':
+	fmtbuf[column] = 'b';
+	column++;
+	break;
+      case '\f':
+	fmtbuf[column] = 'f';
+	column++;
+	break;
+      case '\n':
+	fmtbuf[column] = 'n';
+	column++;
+	break;
+      case '\r':
+	fmtbuf[column] = 'r';
+	column++;
+	break;
+      case '\t':
+	fmtbuf[column] = 't';
+	column++;
+	break;
+      case '\v':
+	fmtbuf[column] = 'v';
+	column++;
+	break;
+      default:
+	i = (c>>6)&03;
+	fmtbuf[column] = i + '0';
+	column++;
+	i = (c>>3)&07;
+	fmtbuf[column] = i + '0';
+	column++;
+	i = (c>>0)&07;
+	fmtbuf[column] = i + '0';
+	column++;
+	break;
+      }
+    }
+  }
+  fmtbuf[column] = '\0';
+  return fmtbuf;
+}
 /* -------------------------------------------------------------------------
     MakeLog has 1 main flaws:
       The message in its entirity, must fit into the tempbuffer.  
@@ -517,23 +597,26 @@ void make_log_entry(enum loglevels loglevel, enum logtypes logtype,
 		    char *message, ...)
 {
   va_list args;
-  char log_buffer[MAXLOGSIZE];
-#ifndef DISABLE_LOGGER
-  char log_details_buffer[MAXLOGSIZE];
-
-  log_file_data_pair *logs;
-#endif
-  
+  char temp_buffer[MAXLOGSIZE];
+  char log_buffer[4*MAXLOGSIZE];
   /* fn is not reentrant but is used in signal handler 
    * with LOGGER it's a little late source name and line number
    * are already changed.
   */
   static int inlog = 0;
 
+#ifndef DISABLE_LOGGER
+  char log_details_buffer[MAXLOGSIZE];
+#ifdef OPEN_LOGS_AS_UID
+  uid_t process_uid;
+#endif
+  log_file_data_pair *logs;
+#endif
+
   if (inlog)
      return;
   inlog = 1;
-
+  
 #ifndef DISABLE_LOGGER
   log_init();
 
@@ -552,10 +635,11 @@ void make_log_entry(enum loglevels loglevel, enum logtypes logtype,
   /* Initialise the Messages */
   va_start(args, message);
 
-  vsnprintf(log_buffer, sizeof(log_buffer), message, args);
+  vsnprintf(temp_buffer, sizeof(temp_buffer), message, args);
 
   /* Finished with args for now */
   va_end(args);
+  format_text(log_buffer, temp_buffer);
 
 #ifdef DISABLE_LOGGER
   syslog(get_syslog_equivalent(loglevel), "%s", log_buffer);
@@ -597,8 +681,15 @@ void make_log_entry(enum loglevels loglevel, enum logtypes logtype,
     if ((*logs)[1].log_file==NULL) {
 #ifdef DEBUG_OUTPUT_TO_SCREEN
       printf("Opening the Log, filename is %s\n", (*logs)[1].log_filename);
+#endif
+#ifdef OPEN_LOGS_AS_UID
+   process_uid = getuid();
+   setuid(OPEN_LOGS_AS_UID);
 #endif
       (*logs)[1].log_file     = fopen((*logs)[1].log_filename, "at");
+#ifdef OPEN_LOGS_AS_UID
+   setuid(process_uid);
+#endif
       if ((*logs)[1].log_file == NULL)
       {
         (*logs)[1].log_file = stdout;
@@ -632,7 +723,9 @@ void make_log_entry(enum loglevels loglevel, enum logtypes logtype,
       printf("Closed\n");
 #endif
     }
-#endif 
+#else // KEEP_LOGFILES_OPEN
+    fflush((*logs)[1].log_file);
+#endif // KEEP_LOGFILES_OPEN
   }
 #endif
 
@@ -720,17 +813,19 @@ void generate_message_details(char *message_details_buffer,
                               struct tag_log_file_data *log_struct,
                               enum loglevels loglevel, enum logtypes logtype)
 {
+#if 0
   char datebuffer[32];
   char processinfo[64];
+  char log_buffer[MAXLOGSIZE];
+  const char *logtype_string;
+
+  char loglevel_string[12]; /* max int size is 2 billion, or 10 digits */
+#endif
 
   char *ptr = message_details_buffer;
   int   templen;
   int   len = message_details_buffer_length;
 
-  char log_buffer[MAXLOGSIZE];
-  const char *logtype_string;
-
-  char loglevel_string[12]; /* max int size is 2 billion, or 10 digits */
 
   struct what_to_print_array what_to_print;
 
@@ -797,7 +892,7 @@ void generate_message_details(char *message_details_buffer,
   if (what_to_print.print_srcfile || what_to_print.print_srcline)
   {
     char sprintf_buffer[8];
-    char *buff_ptr;
+    char *buff_ptr=NULL;
 
     sprintf_buffer[0] = '[';
     if (what_to_print.print_srcfile)
@@ -929,8 +1024,7 @@ void setuplog(char *logtype, char *loglevel, char *filename)
     can be taken from default if needs be.  
    */
   /* const char* sources[] = {"syslog", "filelog"}; */
-  const char *null = "";
-  int sourcenum, typenum, levelnum;
+  int typenum, levelnum;
   log_file_data_pair *logs = log_file_arr[logtype_default];
 
   /*
